@@ -62,6 +62,7 @@ class Build(properties.PropertiesMixin):
 
     VIRTUAL_BUILDERNAME_PROP = "virtual_builder_name"
     VIRTUAL_BUILDERDESCRIPTION_PROP = "virtual_builder_description"
+    VIRTUAL_BUILDER_PROJECT_PROP = "virtual_builder_project"
     VIRTUAL_BUILDERTAGS_PROP = "virtual_builder_tags"
     workdir = "build"
     reason = "changes"
@@ -71,7 +72,10 @@ class Build(properties.PropertiesMixin):
     set_runtime_properties = True
     subs = None
 
-    _sentinel = []  # used as a sentinel to indicate unspecified initial_value
+    class Sentinel:
+        pass
+
+    _sentinel = Sentinel()  # used as a sentinel to indicate unspecified initial_value
 
     def __init__(self, requests):
         self.requests = requests
@@ -95,6 +99,7 @@ class Build(properties.PropertiesMixin):
         # overall results, may downgrade after each step
         self.results = SUCCESS
         self.properties = properties.Properties()
+        self.stopped_reason = None
 
         # tracks execution during the build finish phase
         self._locks_released = False
@@ -157,8 +162,8 @@ class Build(properties.PropertiesMixin):
         return files
 
     def __repr__(self):
-        return "<Build {} number:{} results:{}>".format(self.builder.name, repr(self.number),
-                                                        statusToString(self.results))
+        return (f"<Build {self.builder.name} number:{repr(self.number)} "
+                f"results:{statusToString(self.results)}>")
 
     def blamelist(self):
         # Note that this algorithm is also implemented in
@@ -271,15 +276,19 @@ class Build(properties.PropertiesMixin):
                 description = self.getProperty(
                     self.VIRTUAL_BUILDERDESCRIPTION_PROP,
                     self.builder.config.description)
+                project = self.getProperty(
+                    self.VIRTUAL_BUILDER_PROJECT_PROP,
+                    self.builder.config.project)
                 tags = self.getProperty(
                     self.VIRTUAL_BUILDERTAGS_PROP,
                     self.builder.config.tags)
                 if type(tags) == type([]) and '_virtual_' not in tags:
                     tags.append('_virtual_')
 
-                self.master.data.updates.updateBuilderInfo(self._builderid,
-                                                           description,
-                                                           tags)
+                projectid = yield self.builder.find_project_id(project)
+                # Note: not waiting for updateBuilderInfo to complete
+                self.master.data.updates.updateBuilderInfo(self._builderid, description, None,
+                                                           None, projectid, tags)
 
             else:
                 self._builderid = yield self.builder.getBuilderId()
@@ -300,7 +309,7 @@ class Build(properties.PropertiesMixin):
         self.workername = worker.workername
         self.worker_info = worker.info
 
-        log.msg("{}.startBuild".format(self))
+        log.msg(f"{self}.startBuild")
 
         # TODO: this will go away when build collapsing is implemented; until
         # then we just assign the build to the first buildrequest
@@ -321,9 +330,21 @@ class Build(properties.PropertiesMixin):
         # the preparation step counts the time needed for preparing the worker and getting the
         # locks.
         # we cannot use a real step as we don't have a worker yet.
-        self.preparation_step = buildstep.BuildStep(name="worker_preparation")
+        self.preparation_step = buildstep.create_step_from_step_or_factory(
+            buildstep.BuildStep(name="worker_preparation")
+        )
         self.preparation_step.setBuild(self)
         yield self.preparation_step.addStep()
+
+        # TODO: the time consuming actions during worker preparation are as follows:
+        #  - worker substantiation
+        #  - acquiring build locks
+        # Since locks_acquire_at calculates the time since beginning of the step until the end,
+        # it's impossible to represent this sequence using a single step. In the future it makes
+        # sense to add two steps: one for worker substantiation and another for acquiring build
+        # locks.
+        yield self.master.data.updates.set_step_locks_acquired_at(self.preparation_step.stepid)
+
         Build.setupBuildProperties(self.getProperties(), self.requests,
                                    self.sources, self.number)
 
@@ -383,7 +404,7 @@ class Build(properties.PropertiesMixin):
         # us in a build.
         yield self.master.data.updates.setBuildStateString(self.buildid,
                                                            'pinging worker')
-        log.msg("starting build {}.. pinging the worker {}".format(self, workerforbuilder))
+        log.msg(f"starting build {self}.. pinging the worker {workerforbuilder}")
         try:
             ping_success_or_failure = yield workerforbuilder.ping()
         except Exception:
@@ -416,7 +437,7 @@ class Build(properties.PropertiesMixin):
                                                            'acquiring locks')
         yield self.acquireLocks()
 
-        readymsg = "worker {} ready".format(self.getWorkerName())
+        readymsg = f"worker {self.getWorkerName()} ready"
         yield self.master.data.updates.setStepStateString(self.preparation_step.stepid, readymsg)
         yield self.master.data.updates.finishStep(self.preparation_step.stepid, SUCCESS, False)
 
@@ -439,6 +460,8 @@ class Build(properties.PropertiesMixin):
             self.workerforbuilder.worker.putInQuarantine()
             if isinstance(why, failure.Failure):
                 yield self.preparation_step.addLogWithFailure(why)
+            elif isinstance(why, Exception):
+                yield self.preparation_step.addLogWithException(why)
             yield self.master.data.updates.setStepStateString(self.preparation_step.stepid,
                                                             "error while " + state_string)
             yield self.master.data.updates.finishStep(self.preparation_step.stepid,
@@ -459,10 +482,10 @@ class Build(properties.PropertiesMixin):
             return defer.succeed(None)
         if self.stopped:
             return defer.succeed(None)
-        log.msg("acquireLocks(build {}, locks {})".format(self, self.locks))
+        log.msg(f"acquireLocks(build {self}, locks {self.locks})")
         for lock, access in self.locks:
             if not lock.isAvailable(self, access):
-                log.msg("Build {} waiting for lock {}".format(self, lock))
+                log.msg(f"Build {self} waiting for lock {lock}")
                 d = lock.waitUntilMaybeAvailable(self, access)
                 d.addCallback(self.acquireLocks)
                 self._acquiringLock = (lock, access, d)
@@ -480,7 +503,7 @@ class Build(properties.PropertiesMixin):
             count = self.stepnames[name]
             count += 1
             self.stepnames[name] = count
-            name = "{}_{}".format(step.name, count)
+            name = f"{step.name}_{count}"
         else:
             self.stepnames[name] = 0
         step.name = name
@@ -568,7 +591,7 @@ class Build(properties.PropertiesMixin):
             yield self.startNextStep()
 
         except Exception as e:
-            log.msg("{} build got exception when running step {}".format(self, step))
+            log.msg(f"{self} build got exception when running step {step}")
             log.err(e)
 
             yield self.master.data.updates.setBuildProperties(self.buildid, self)
@@ -586,11 +609,11 @@ class Build(properties.PropertiesMixin):
         text = None
         if isinstance(results, tuple):
             results, text = results
-        assert isinstance(results, type(SUCCESS)), "got %r" % (results,)
+        assert isinstance(results, type(SUCCESS)), f"got {repr(results)}"
         summary = yield step.getBuildResultSummary()
         if 'build' in summary:
             text = [summary['build']]
-        log.msg(" step '{}' complete: {} ({})".format(step.name, statusToString(results), text))
+        log.msg(f" step '{step.name}' complete: {statusToString(results)} ({text})")
         if text:
             self.text.extend(text)
             self.master.data.updates.setBuildStateString(self.buildid,
@@ -607,7 +630,7 @@ class Build(properties.PropertiesMixin):
         # the worker went away. There are several possible reasons for this,
         # and they aren't necessarily fatal. For now, kill the build, but
         # TODO: see if we can resume the build when it reconnects.
-        log.msg("{}.lostRemote".format(self))
+        log.msg(f"{self}.lostRemote")
         self.conn = None
         self.text = ["lost", "connection"]
         self.results = RETRY
@@ -625,6 +648,7 @@ class Build(properties.PropertiesMixin):
     def controlStopBuild(self, key, params):
         return self.stopBuild(**params)
 
+    @defer.inlineCallbacks
     def stopBuild(self, reason="<no reason given>", results=CANCELLED):
         # the idea here is to let the user cancel a build because, e.g.,
         # they realized they committed a bug and they don't want to waste
@@ -633,13 +657,13 @@ class Build(properties.PropertiesMixin):
         # build as failed quickly rather than waiting for the worker's
         # timeout to kill it on its own.
 
-        log.msg(" {}: stopping build: {} {}".format(self, reason, results))
+        log.msg(f" {self}: stopping build: {reason} {results}")
         if self.finished:
             return
-        # TODO: include 'reason' in this point event
+        self.stopped_reason = reason
         self.stopped = True
         if self.currentStep and self.currentStep.results is None:
-            self.currentStep.interrupt(reason)
+            yield self.currentStep.interrupt(reason)
 
         self.results = results
 
@@ -664,6 +688,8 @@ class Build(properties.PropertiesMixin):
             text = ["cancelled"]
         else:
             text = ["build", "successful"]
+        if self.stopped_reason is not None:
+            text.extend([f'({self.stopped_reason})'])
         text.extend(self.text)
         return self.buildFinished(text, self.results)
 
@@ -688,7 +714,7 @@ class Build(properties.PropertiesMixin):
                 self.subs.unsubscribe()
                 self.subs = None
                 self.conn = None
-            log.msg(" {}: build finished".format(self))
+            log.msg(f" {self}: build finished")
             self.results = worst_status(self.results, results)
             eventually(self.releaseLocks)
             metrics.MetricCountEvent.log('active_builds', -1)
@@ -719,7 +745,7 @@ class Build(properties.PropertiesMixin):
 
     def releaseLocks(self):
         if self.locks:
-            log.msg("releaseLocks({}): {}".format(self, self.locks))
+            log.msg(f"releaseLocks({self}): {self.locks}")
 
         for lock, access in self.locks:
             if lock.isOwner(self, access):
