@@ -13,20 +13,24 @@
 #
 # Copyright Buildbot Team Members
 
+
 import re
 import textwrap
 
-from twisted.internet import defer, reactor
+from twisted.internet import defer
+from twisted.internet import reactor
+from zope.interface import implementer
 
 from buildbot import util
-from buildbot.process import buildstep
-from buildbot.steps.source.base import Source
 from buildbot.interfaces import IRenderable
-from zope.interface import implements
+from buildbot.process import buildstep
+from buildbot.process import remotecommand
+from buildbot.process import results
+from buildbot.steps.source.base import Source
 
 
-class RepoDownloadsFromProperties(util.ComparableMixin, object):
-    implements(IRenderable)
+@implementer(IRenderable)
+class RepoDownloadsFromProperties(util.ComparableMixin):
     parse_download_re = (re.compile(r"repo download ([^ ]+) ([0-9]+/[0-9]+)"),
                          re.compile(r"([^ ]+) ([0-9]+/[0-9]+)"),
                          re.compile(r"([^ ]+)/([0-9]+/[0-9]+)"),
@@ -48,10 +52,11 @@ class RepoDownloadsFromProperties(util.ComparableMixin, object):
     def parseDownloadProperty(self, s):
         """
          lets try to be nice in the format we want
-         can support several instances of "repo download proj number/patch" (direct copy paste from gerrit web site)
-         or several instances of "proj number/patch" (simpler version)
-         This feature allows integrator to build with several pending interdependant changes.
-         returns list of repo downloads sent to the buildslave
+         can support several instances of "repo download proj number/patch"
+         (direct copy paste from gerrit web site) or several instances of "proj number/patch"
+         (simpler version)
+         This feature allows integrator to build with several pending interdependent changes.
+         returns list of repo downloads sent to the worker
          """
         if s is None:
             return []
@@ -59,14 +64,14 @@ class RepoDownloadsFromProperties(util.ComparableMixin, object):
         for cur_re in self.parse_download_re:
             res = cur_re.search(s)
             while res:
-                ret.append("%s %s" % (res.group(1), res.group(2)))
+                ret.append(f"{res.group(1)} {res.group(2)}")
                 s = s[:res.start(0)] + s[res.end(0):]
                 res = cur_re.search(s)
         return ret
 
 
-class RepoDownloadsFromChangeSource(util.ComparableMixin, object):
-    implements(IRenderable)
+@implementer(IRenderable)
+class RepoDownloadsFromChangeSource(util.ComparableMixin):
     compare_attrs = ('codebase',)
 
     def __init__(self, codebase=None):
@@ -80,29 +85,32 @@ class RepoDownloadsFromChangeSource(util.ComparableMixin, object):
             changes = props.getBuild().getSourceStamp(self.codebase).changes
         for change in changes:
             if ("event.type" in change.properties and
-               change.properties["event.type"] == "patchset-created"):
-                    downloads.append("%s %s/%s" % (change.properties["event.change.project"],
-                                                   change.properties["event.change.number"],
-                                                   change.properties["event.patchSet.number"]))
+                    change.properties["event.type"] == "patchset-created"):
+                downloads.append(f'{change.properties["event.change.project"]} '
+                                 f'{change.properties["event.change.number"]}/'
+                                 f'{change.properties["event.patchSet.number"]}')
         return downloads
 
 
 class Repo(Source):
+
     """ Class for Repo with all the smarts """
     name = 'repo'
-    renderables = ["manifestURL", "manifestFile", "tarball", "jobs",
+    renderables = ["manifestURL", "manifestBranch", "manifestFile", "tarball", "jobs",
                    "syncAllBranches", "updateTarballAge", "manifestOverrideUrl",
-                   "repoDownloads"]
+                   "repoDownloads", "depth", "submodules"]
 
     ref_not_found_re = re.compile(r"fatal: Couldn't find remote ref")
     cherry_pick_error_re = re.compile(r"|".join([r"Automatic cherry-pick failed",
-                                                r"error: "
-                                                r"fatal: "
-                                                r"possibly due to conflict resolution."]))
+                                                 r"error: "
+                                                 r"fatal: "
+                                                 r"possibly due to conflict resolution."]))
     re_change = re.compile(r".* refs/changes/\d\d/(\d+)/(\d+) -> FETCH_HEAD$")
     re_head = re.compile(r"^HEAD is now at ([0-9a-f]+)...")
-    mirror_sync_retry = 10  # number of retries, if we detect mirror desynchronization
-    mirror_sync_sleep = 60  # wait 1min between retries (thus default total retry time is 10min)
+    # number of retries, if we detect mirror desynchronization
+    mirror_sync_retry = 10
+    # wait 1min between retries (thus default total retry time is 10min)
+    mirror_sync_sleep = 60
 
     def __init__(self,
                  manifestURL=None,
@@ -111,9 +119,12 @@ class Repo(Source):
                  tarball=None,
                  jobs=None,
                  syncAllBranches=False,
-                 updateTarballAge=7*24.0*3600.0,
+                 updateTarballAge=7 * 24.0 * 3600.0,
                  manifestOverrideUrl=None,
                  repoDownloads=None,
+                 depth=0,
+                 submodules=False,
+                 syncQuietly=False,
                  **kwargs):
         """
         @type  manifestURL: string
@@ -141,6 +152,15 @@ class Repo(Source):
         @type repoDownloads: list of strings
         @param repoDownloads: optional repo download to perform after the repo sync
 
+        @type depth: integer
+        @param depth: optional depth parameter to repo init.
+                          If specified, create a shallow clone with given depth.
+
+        @type submodules: string
+        @param submodules: optional submodules parameter to repo init.
+
+        @type syncQuietly: bool.
+        @param syncQuietly: true, then suppress verbose output from repo sync.
         """
         self.manifestURL = manifestURL
         self.manifestBranch = manifestBranch
@@ -153,7 +173,10 @@ class Repo(Source):
         if repoDownloads is None:
             repoDownloads = []
         self.repoDownloads = repoDownloads
-        Source.__init__(self, **kwargs)
+        self.depth = depth
+        self.submodules = submodules
+        self.syncQuietly = syncQuietly
+        super().__init__(**kwargs)
 
         assert self.manifestURL is not None
 
@@ -174,10 +197,10 @@ class Repo(Source):
         manifest_related_downloads = []
         for download in self.repoDownloads:
             project, ch_ps = download.split(" ")[-2:]
-            if (self.manifestURL.endswith("/"+project) or
-               self.manifestURL.endswith("/"+project+".git")):
+            if (self.manifestURL.endswith("/" + project) or
+                    self.manifestURL.endswith("/" + project + ".git")):
                 ch, ps = map(int, ch_ps.split("/"))
-                branch = "refs/changes/%02d/%d/%d" % (ch % 100, ch, ps)
+                branch = f"refs/changes/{ch % 100:02}/{ch}/{ps}"
                 manifest_related_downloads.append(
                     ["git", "fetch", self.manifestURL, branch])
                 manifest_related_downloads.append(
@@ -188,30 +211,31 @@ class Repo(Source):
         self.manifestDownloads = manifest_related_downloads
 
     def _repoCmd(self, command, abandonOnFailure=True, **kwargs):
-        return self._Cmd(["repo"]+command, abandonOnFailure=abandonOnFailure, **kwargs)
+        return self._Cmd(["repo"] + command, abandonOnFailure=abandonOnFailure, **kwargs)
 
+    @defer.inlineCallbacks
     def _Cmd(self, command, abandonOnFailure=True, workdir=None, **kwargs):
         if workdir is None:
             workdir = self.workdir
-        self.cmd = cmd = buildstep.RemoteShellCommand(workdir, command,
-                                                      env=self.env,
-                                                      logEnviron=self.logEnviron,
-                                                      timeout=self.timeout,  **kwargs)
+        cmd = remotecommand.RemoteShellCommand(workdir, command,
+                                               env=self.env,
+                                               logEnviron=self.logEnviron,
+                                               timeout=self.timeout, **kwargs)
+        self.lastCommand = cmd
         # does not make sense to logEnviron for each command (just for first)
         self.logEnviron = False
         cmd.useLog(self.stdio_log, False)
-        self.stdio_log.addHeader("Starting command: %s\n" % (" ".join(command), ))
-        self.step_status.setText(["%s" % (" ".join(command[:2]))])
-        d = self.runCommand(cmd)
+        yield self.stdio_log.addHeader(f'Starting command: {" ".join(command)}\n')
+        self.description = ' '.join(command[:2])
+        # FIXME: enable when new style step is switched on yield self.updateSummary()
+        yield self.runCommand(cmd)
 
-        def evaluateCommand(cmd):
-            if abandonOnFailure and cmd.didFail():
-                self.step_status.setText(["repo failed at: %s" % (" ".join(command[:2]))])
-                self.stdio_log.addStderr("Source step failed while running command %s\n" % cmd)
-                raise buildstep.BuildStepFailed()
-            return cmd.rc
-        d.addCallback(lambda _: evaluateCommand(cmd))
-        return d
+        if abandonOnFailure and cmd.didFail():
+            self.descriptionDone = f'repo failed at: {" ".join(command[:2])}'
+            msg = f"Source step failed while running command {cmd}\n"
+            yield self.stdio_log.addStderr(msg)
+            raise buildstep.BuildStepFailed()
+        return cmd.rc
 
     def repoDir(self):
         return self.build.path_module.join(self.workdir, ".repo")
@@ -219,39 +243,35 @@ class Repo(Source):
     def sourcedirIsUpdateable(self):
         return self.pathExists(self.repoDir())
 
-    def startVC(self, branch, revision, patch):
-        d = self.doStartVC()
-        d.addErrback(self.failed)
+    def run_vc(self, branch, revision, patch):
+        return self.doStartVC()
 
     @defer.inlineCallbacks
     def doStartVC(self):
-        self.stdio_log = self.addLogForRemoteCommands("stdio")
+        self.stdio_log = yield self.addLogForRemoteCommands("stdio")
 
         self.filterManifestPatches()
 
         if self.repoDownloads:
-            self.stdio_log.addHeader("will download:\n" + "repo download " + "\nrepo download ".join(self.repoDownloads) + "\n")
+            yield self.stdio_log.addHeader("will download:\nrepo download {}\n".format(
+                    "\nrepo download ".join(self.repoDownloads)))
 
         self.willRetryInCaseOfFailure = True
 
-        d = self.doRepoSync()
+        try:
+            yield self.doRepoSync()
+        except buildstep.BuildStepFailed as e:
+            if not self.willRetryInCaseOfFailure:
+                raise
+            yield self.stdio_log.addStderr("got issue at first try:\n" + str(e) +
+                                           "\nRetry after clobber...")
+            yield self.doRepoSync(forceClobber=True)
 
-        def maybeRetry(why):
-            # in case the tree was corrupted somehow because of previous build
-            # we clobber one time, and retry everything
-            if why.check(buildstep.BuildStepFailed) and self.willRetryInCaseOfFailure:
-                self.stdio_log.addStderr("got issue at first try:\n" + str(why) +
-                                         "\nRetry after clobber...")
-                return self.doRepoSync(forceClobber=True)
-            return why  # propagate to self.failed
-        d.addErrback(maybeRetry)
-        yield d
         yield self.maybeUpdateTarball()
 
         # starting from here, clobbering will not help
         yield self.doRepoDownloads()
-        self.setStatus(self.cmd, 0)
-        yield self.finished(0)
+        return results.SUCCESS
 
     @defer.inlineCallbacks
     def doClobberStart(self):
@@ -267,15 +287,23 @@ class Repo(Source):
             self.willRetryInCaseOfFailure = False
             yield self.doClobberStart()
         yield self.doCleanup()
-        yield self._repoCmd(['init',
-                             '-u', self.manifestURL,
-                             '-b', self.manifestBranch,
-                             '-m', self.manifestFile])
+        command = ['init',
+                   '-u', self.manifestURL,
+                   '-b', self.manifestBranch,
+                   '-m', self.manifestFile,
+                   '--depth', str(self.depth)]
+
+        if self.submodules:
+            command.append('--submodules')
+
+        yield self._repoCmd(command)
 
         if self.manifestOverrideUrl:
-            self.stdio_log.addHeader("overriding manifest with %s\n" % (self.manifestOverrideUrl))
-            local_file = yield self.pathExists(self.build.path_module.join(self.workdir,
-                                                                          self.manifestOverrideUrl))
+            msg = f"overriding manifest with {self.manifestOverrideUrl}\n"
+            yield self.stdio_log.addHeader(msg)
+
+            local_path = self.build.path_module.join(self.workdir, self.manifestOverrideUrl)
+            local_file = yield self.pathExists(local_path)
             if local_file:
                 yield self._Cmd(["cp", "-f", self.manifestOverrideUrl, "manifest_override.xml"])
             else:
@@ -284,16 +312,20 @@ class Repo(Source):
                             workdir=self.build.path_module.join(self.workdir, ".repo"))
 
         for command in self.manifestDownloads:
-            yield self._Cmd(command, workdir=self.build.path_module.join(self.workdir, ".repo", "manifests"))
+            yield self._Cmd(command, workdir=self.build.path_module.join(self.workdir, ".repo",
+                                                                         "manifests"))
 
-        command = ['sync']
+        command = ['sync', '--force-sync']
         if self.jobs:
             command.append('-j' + str(self.jobs))
         if not self.syncAllBranches:
             command.append('-c')
-        self.step_status.setText(["repo sync"])
-        self.stdio_log.addHeader("synching manifest %s from branch %s from %s\n"
-                                 % (self.manifestFile, self.manifestBranch, self.manifestURL))
+        if self.syncQuietly:
+            command.append('-q')
+        self.description = "repo sync"
+        # FIXME: enable when new style step is used: yield self.updateSummary()
+        yield self.stdio_log.addHeader(f"synching manifest {self.manifestFile} from branch "
+                                       f"{self.manifestBranch} from {self.manifestURL}\n")
         yield self._repoCmd(command)
 
         command = ['manifest', '-r', '-o', 'manifest-original.xml']
@@ -303,10 +335,10 @@ class Repo(Source):
     # compiled regexps in self.re_error_messages
     def _findErrorMessages(self, error_re):
         for logname in ['stderr', 'stdout']:
-            if not hasattr(self.cmd, logname):
+            if not hasattr(self.lastCommand, logname):
                 continue
-            msg = getattr(self.cmd, logname)
-            if not (re.search(error_re, msg) is None):
+            msg = getattr(self.lastCommand, logname)
+            if not re.search(error_re, msg) is None:
                 return True
         return False
 
@@ -320,8 +352,7 @@ class Repo(Source):
         self.repo_downloaded = ""
         for download in self.repoDownloads:
             command = ['download'] + download.split(' ')
-            self.stdio_log.addHeader("downloading changeset %s\n"
-                                     % (download))
+            yield self.stdio_log.addHeader(f"downloading changeset {download}\n")
 
             retry = self.mirror_sync_retry + 1
             while retry > 0:
@@ -330,25 +361,24 @@ class Repo(Source):
                 if not self._findErrorMessages(self.ref_not_found_re):
                     break
                 retry -= 1
-                self.stdio_log.addStderr("failed downloading changeset %s\n" % (download))
-                self.stdio_log.addHeader("wait one minute for mirror sync\n")
+                yield self.stdio_log.addStderr(f"failed downloading changeset {download}\n")
+                yield self.stdio_log.addHeader("wait one minute for mirror sync\n")
                 yield self._sleep(self.mirror_sync_sleep)
 
             if retry == 0:
-                self.step_status.setText(["repo: change %s does not exist" % download])
-                self.step_status.setText2(["repo: change %s does not exist" % download])
+                self.descriptionDone = f"repo: change {download} does not exist"
                 raise buildstep.BuildStepFailed()
 
-            if self.cmd.didFail() or self._findErrorMessages(self.cherry_pick_error_re):
+            if self.lastCommand.didFail() or self._findErrorMessages(self.cherry_pick_error_re):
                 # cherry pick error! We create a diff with status current workdir
                 # in stdout, which reveals the merge errors and exit
                 command = ['forall', '-c', 'git', 'diff', 'HEAD']
                 yield self._repoCmd(command, abandonOnFailure=False)
-                self.step_status.setText(["download failed: %s" % download])
+                self.descriptionDone = f"download failed: {download}"
                 raise buildstep.BuildStepFailed()
 
-            if hasattr(self.cmd, 'stderr'):
-                lines = self.cmd.stderr.split("\n")
+            if hasattr(self.lastCommand, 'stderr'):
+                lines = self.lastCommand.stderr.split("\n")
                 match1 = match2 = False
                 for line in lines:
                     if not match1:
@@ -356,9 +386,8 @@ class Repo(Source):
                     if not match2:
                         match2 = self.re_head.match(line)
                 if match1 and match2:
-                    self.repo_downloaded += "%s/%s %s " % (match1.group(1),
-                                                           match1.group(2),
-                                                           match2.group(1))
+                    self.repo_downloaded += (f"{match1.group(1)}/{match1.group(2)} "
+                                             f"{match2.group(1)} ")
 
         self.setProperty("repo_downloaded", self.repo_downloaded, "Source")
 
@@ -366,13 +395,16 @@ class Repo(Source):
         # Keep in mind that the compression part of tarball generation
         # can be non negligible
         tar = ['tar']
-        if self.tarball.endswith("gz"):
+        if self.tarball.endswith("pigz"):
+            tar.append('-I')
+            tar.append('pigz')
+        elif self.tarball.endswith("gz"):
             tar.append('-z')
-        if self.tarball.endswith("bz2") or self.tarball.endswith("bz"):
+        elif self.tarball.endswith("bz2") or self.tarball.endswith("bz"):
             tar.append('-j')
-        if self.tarball.endswith("lzma"):
+        elif self.tarball.endswith("lzma"):
             tar.append('--lzma')
-        if self.tarball.endswith("lzop"):
+        elif self.tarball.endswith("lzop"):
             tar.append('--lzop')
         return tar
 
@@ -389,23 +421,25 @@ class Repo(Source):
     def maybeUpdateTarball(self):
         if not self.tarball or self.updateTarballAge is None:
             return
-        # tarball path is absolute, so we cannot use slave's stat command
+        # tarball path is absolute, so we cannot use worker's stat command
         # stat -c%Y gives mtime in second since epoch
-        res = yield self._Cmd(["stat", "-c%Y", self.tarball], collectStdout=True, abandonOnFailure=False)
+        res = yield self._Cmd(["stat", "-c%Y", self.tarball], collectStdout=True,
+                              abandonOnFailure=False)
         if not res:
-            tarball_mtime = int(self.cmd.stdout)
+            tarball_mtime = int(self.lastCommand.stdout)
             yield self._Cmd(["stat", "-c%Y", "."], collectStdout=True)
-            now_mtime = int(self.cmd.stdout)
+            now_mtime = int(self.lastCommand.stdout)
             age = now_mtime - tarball_mtime
         if res or age > self.updateTarballAge:
-            tar = self.computeTarballOptions() + ['-cvf', self.tarball, ".repo"]
+            tar = self.computeTarballOptions() + \
+                ['-cvf', self.tarball, ".repo"]
             res = yield self._Cmd(tar, abandonOnFailure=False)
-            if res:  # error with tarball.. erase tarball, but dont fail
+            if res:  # error with tarball.. erase tarball, but don't fail
                 yield self._Cmd(["rm", "-f", self.tarball], abandonOnFailure=False)
 
     # a simple shell script to gather all cleanup tweaks...
     # doing them one by one just complicate the stuff
-    # and messup the stdio log
+    # and mess up the stdio log
     def _getCleanupCommand(self):
         """also used by tests for expectations"""
         return textwrap.dedent("""\
@@ -427,7 +461,9 @@ class Repo(Source):
              repo forall -c git clean -f -d -x 2>/dev/null
              repo forall -c git reset --hard HEAD 2>/dev/null
              rm -f %(workdir)s/.repo/project.list
-             """) % self.__dict__
+             """) % {"manifestBranch": self.manifestBranch,
+                     "manifestFile": self.manifestFile,
+                     "workdir": self.workdir}
 
     def doCleanup(self):
         command = self._getCleanupCommand()
